"""
Filters out doublets using scrublet.
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""
# external package imports
from anndata import AnnData
import matplotlib.pyplot as plt
import scrublet as scr
import pandas as pd

# scrnatools package imports
from .._configs import configs

logger = configs.create_logger(__name__.split('_', 1)[1])

# -------------------------------------------------------function----------------------------------------------------- #


def scrublet(
        adata: AnnData,
        raw_counts_layer: str,
        doublet_threshold: float = 0.2,
        batch_key: str = None,
        subset: bool = False,
) -> AnnData:
    """Filters out doublets using scrublet.

    Args:
        adata (AnnData): The dataset to process
        raw_counts_layer (str):  A key in 'adata.layers' pointing to the raw counts data.
        doublet_threshold (float, optional): The doublet score threshold to call doublets/singlets on. Defaults to 0.2.
        batch_key (str, optional): A column in 'adata.obs' annotating different batches of data. If 'None' treats all cells as coming from the same batch. Defaults to None.
        subset (bool, optional): Whether to filter the doublets in place or just annotate them.

    Returns:
        AnnData: The dataset provided with doublets filtered/labeled based on doublet_threshold
    """

    # Setup
    scrublet_predictions = pd.DataFrame()
    if batch_key is None:
        adata.obs["scrublet_batch_key"] = "None"
        batch_key = "scrublet_batch_key"
    # Calculate doublet scores for each batch indepenently
    plt.rcParams["axes.grid"] = False
    for key in adata.obs[batch_key].unique():
        subset_data = adata[adata.obs[batch_key] == key]
        counts_matrix = subset_data.layers[raw_counts_layer]
        scrub = scr.Scrublet(counts_matrix, expected_doublet_rate=0.06)
        doublet_scores, predicted_doublets = scrub.scrub_doublets(
            min_counts=2,
            min_cells=3,
            min_gene_variability_pctl=85,
            n_prin_comps=30
        )
        scrub.call_doublets(threshold=doublet_threshold)
        scrub.plot_histogram()
        scrublet_prediction = pd.DataFrame(
            {"scrublet_score": doublet_scores},
            index=subset_data.obs.index
        )
        scrublet_predictions = pd.concat(
            [scrublet_predictions, scrublet_prediction]
        )
    # Add scrublet scores to adata.obs and filter out cells with a score > doublet_threshold
    adata.obs["scrublet_score"] = scrublet_predictions.scrublet_score
    adata.obs["scrublet_called_doublet"] = adata.obs.scrublet_score > doublet_threshold
    num_doublets = adata.obs.scrublet_called_doublet.sum()
    pct_doublets = round(num_doublets / len(adata) * 100, 3)
    logger.info(
        f"{pct_doublets}% of cells classified as doublets ({num_doublets} cells)")
    if subset:
        adata = adata[~adata.obs.scrublet_called_doublet]
    return adata



"""
Plots the density of expression of genes on an embedding.
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""
# external package imports
from anndata import AnnData
from typing import List, Tuple
import seaborn as sns
import matplotlib.pyplot as plt
from math import ceil
from scipy.sparse import issparse
from statsmodels.nonparametric.kernel_density import KDEMultivariateConditional, EstimatorSettings
from sklearn.preprocessing import StandardScaler
import numpy as np

# scrnatools package imports
from .._configs import configs
from .._utils import check_path

logger = configs.create_logger(__name__.split('_', 1)[1])

# -------------------------------------------------------function----------------------------------------------------- #


def gene_density_plot(
        adata: AnnData,
        gene_list: List[str],
        data_loc: str = "X",
        thresh: int = 1,
        latent_rep: str = "X_umap",
        est_settings=None,
        cmap: str = "magma",
        s: int = None,
        ncols: int = 3,
        figsize: Tuple[int] = (3, 3),
        title: str = None,
        save_path: str = None,
        dpi: int = 300,
):
    """Plots the density of expression of genes on an embedding.

    Args:
        adata (AnnData): The dataset containing the gene expression and cell data.
        gene_list (List[str]): A list of genes to plot.
        data_loc (str, optional): The location of the expression data to use for density calculations, can be a layer in 'adata.layers' or 'X' to use the data stored in adata.X. Defaults to "X".
        thresh (int, optional): Kernel density threshold. Defaults to 1.
        latent_rep (str, optional): The 2D representation to plot gene expression for each cell on. Defaults to "X_umap".
        est_settings (optional): Custom settings for the kernel density estimator. Defaults to None.
        cmap (str, optional): The pyplot colormap to use for plotting. Defaults to "magma".
        s (int, optional): The size of data points to plot. Defaults to None.
        ncols (int, optional): The number of columns in the figure. Defaults to 3.
        figsize (Tuple[int], optional): The dimensions of each subplot for a single gene. Defaults to (3, 3).
        title (str, optional): The title of the figure. Defaults to None.
        save_path (str, optional): The path to save the figure. Defaults to None.
        dpi (int, optional): The resolution of the saved image. Defaults to 300.

    Raises:
        ValueError: If the 'data_loc' provided is not 'X' or a valid layer in 'adata.layers'
    """

    # set up figure
    if len(gene_list) < ncols:
        ncols = len(gene_list)
    nrows = ceil((len(gene_list)) / ncols)
    fig = plt.figure(figsize=(ncols * figsize[0], nrows * figsize[1]))
    sns.set_theme(context="paper", style="white", )

    # Iterate over genes to plot
    for index, gene in enumerate(gene_list):
        if gene not in adata.var_names:
            logger.info(f"{gene} not in 'adata.var_names', skipping")
        else:
            temp_data = adata[:, gene]
            if data_loc == "X":
                if issparse(temp_data.X):
                    x = temp_data.X.todense()
                else:
                    x = temp_data.X
            else:
                if data_loc in temp_data.layers:
                    if issparse(temp_data.X):
                        x = temp_data.layers[data_loc].todense()
                    else:
                        x = temp_data.layers[data_loc]
                else:
                    raise ValueError(
                        f"{data_loc} not 'X' or a valid layer in 'adata.layers'")
            scaler = StandardScaler()
            x_scaled = scaler.fit_transform(x).reshape(-1)
            density = KDEMultivariateConditional(
                endog=x_scaled,
                exog=adata.obsm[latent_rep],
                dep_type="c",
                indep_type="cc",
                bw="normal_reference",
                defaults=est_settings
            )
            z1 = density.cdf(
                thresh + np.zeros_like(x_scaled),
                adata.obsm[latent_rep]
            )

            cm = plt.get_cmap(cmap)
            s = int(80000 / adata.shape[0]) if s is None else s

            # Plot density of gene expression
            idx = np.argsort(x_scaled)
            plt.subplot(nrows, ncols, index + 1)
            ax = sns.scatterplot(
                x=adata.obsm[latent_rep][idx, 0],
                y=adata.obsm[latent_rep][idx, 1],
                hue=1 - z1[idx],
                palette=cm,
                s=s,
                linewidth=0
            )
            ax.set_aspect("equal")  # Make sure the aspect ratio is square
            ax.set(xticklabels=[], yticklabels=[], xlabel=None,
                   ylabel=None, title=gene)  # Get rid of x and y labels

            norm = plt.Normalize(1 - z1[idx].min(), 1 - z1[idx].max())
            sm = plt.cm.ScalarMappable(cmap=cm, norm=norm)
            sm.set_array([])

            # Remove the legend and add a colorbar
            ax.get_legend().remove()
            color_bar = ax.figure.colorbar(sm)
            color_bar.set_label("Probability of Expression")
    sns.despine(left=True, bottom=True)  # no borders
    fig.suptitle(title)  # figure title
    fig.tight_layout(rect=[0, 0.03, 1, 0.98])

    if save_path is not None:
        if "/" not in save_path:
            save_path = f"./{save_path}"
        check_path(save_path.rsplit("/", 1)[0])
        logger.info(f"Saving figure to {save_path}")
        plt.savefig(save_path, dpi=dpi, facecolor="white",)
    plt.show()



"""
Plots violinplot subplots of expression of genes.
From scrnatools package

Created on Wed March 29 17:28:15 2023

@author: joe germino (joe.germino@ucsf.edu), nolan horner (nolan.horner@ucsf.edu)
"""
# external package imports
from anndata import AnnData
from typing import List, Tuple
import seaborn as sns
import matplotlib.pyplot as plt
import math
import scrnatools

# scrnatools package imports
from .._configs import configs
from .._utils import check_path

logger = configs.create_logger(__name__.split('_', 1)[1])

# -------------------------------------------------------function----------------------------------------------------- #


def gene_violinplot(
    adata: AnnData,
    gene_list: List[str],
    x_key: str,
    layer: str = "X",
    x_values: List[str] = ['All'],
    hue_key: str = None,
    hue_values: List[str] = ['All'],
    ncols: int = None,
    nrows: int = None,
    save_path: str = None,
    dpi: int = 300,
    fig_size: Tuple[float] = (0.5, 2.5),
    *args, **kwargs
):
    """Plots violinplot subplots of expression of genes.

    Args:
        adata (AnnData): The dataset containing the gene expression and cell data.
        gene_list (List[str]): A list of genes to plot.
        x_key (str): The categorical grouping to display on the x axis of the violinplot.
        layer (str, optional): The layer containing expression data to use for the violinplot, can be a layer in 'adata.layers' or 'X' to use the data stored in adata.X. Defaults to "X".
        x_values (List[str], optional): Values from x key group to display on violinplot. Defaults to ['All'].
        hue_key (str, optional): The categorical grouping to color the grouped violin plots by. Values will appear in legend. If 'hue_key' is 'None' no hue splitting of x values will occur. Defaults to None.
        hue_values (List[str], optional): Values from hue_key to display on violinplot. Defaults to ['All'].
        ncols (int, optional): Number of columns to display the violinplots. Defaults to None.
        nrows (int, optional): Number of rows to display the violinplots. Defaults to None.
        save_path (str, optional): The path to save the figure. Defaults to None.
        dpi (int, optional): The resolution of the saved image. Defaults to 300.
        fig_size (Tuple[float], optional): The scaling factors for the column and row size in the figure. Defaults to (0.5, 2.5).

    Raises:
        ValueError: If a gene in 'gene_list' provided is not in provided AnnData layer.
        ValueError: If the 'hue_values' is provided but hue_key is not provided.
        ValueError: If the 'hue_values' provided less than 2 if hue_key is also provided.
        ValueError: If ncols * nrows is less than length of gene list provided.
        ValueError: If a value in 'x_values' provided is not in AnnData layer x key.
        ValueError: If a value in 'hue_values' provided is not in AnnData layer hue key given hue key is also provided.
    """

    # check inputs
    invalid_genes = [i for i in gene_list if i not in adata.var_names]
    if len(invalid_genes) > 0:
        raise ValueError(f"Genes not found in data: {invalid_genes}")
    if hue_key is None:
        if hue_values != ['All']:
            raise ValueError(f"Please enter a hue_key.")
    else:
        if len(hue_values) < 2 and hue_values != ['All']:
            raise ValueError(
                f"Please choose more than one hue value. Or use '[All]' to see all {hue_key}s"
            )
    if x_values == ['All']:
        x_values = adata.obs[x_key].unique()
    if hue_values == ['All'] and hue_key != None:
        hue_values = adata.obs[hue_key].unique()

    # format figure
    max_input_len = max(len(x_values), len(hue_values))
    min_input_len = min(len(hue_values), len(x_values))
    if (ncols is None) & (nrows is None):
        if max_input_len > 5:
            ncols = 1
            nrows = len(gene_list)
        elif max_input_len > 2:
            if min_input_len <= 2:
                ncols = 3
            else:
                ncols = 2
        else:
            if min_input_len <= 2:
                ncols = 5
            else:
                ncols = 3
        if (len(gene_list) % ncols) > 0:
            nrows = int((len(gene_list) / ncols)) + 1
        else:
            nrows = int((len(gene_list) / ncols))

    elif (ncols is None) & (nrows is not None):
        if (len(gene_list) % nrows) > 0:
            ncols = int(len(gene_list) / nrows) + 1
        else:
            ncols = int((len(gene_list) / nrows))
    elif (nrows is None) & (ncols is not None):
        if (len(gene_list) % ncols) > 0:
            nrows = int((len(gene_list) / ncols)) + 1
        else:
            nrows = int((len(gene_list) / ncols))

    if ncols * nrows < len(gene_list):
        raise ValueError(
            "Number of rows and columns must fit number of genes in gene list (nrows * ncols >= length of gene list).")

    fig = plt.figure(figsize=(ncols*len(x_values) *
                     max(2, len(hue_values))*fig_size[0], nrows*fig_size[1]))

    # subset data on x values and hue values -> expression matrix
    if hue_key is None:
        subsetAdata = adata[adata.obs[x_key].isin(x_values), gene_list]
    else:
        subsetAdata = adata[adata.obs[x_key].isin(
            x_values) & adata.obs[hue_key].isin(hue_values), gene_list]
    expression_matrix = scrnatools.tl.get_expression_matrix(
        subsetAdata, gene_data=layer)

    expression_matrix[x_key] = subsetAdata.obs[x_key]
    if hue_key != None:
        expression_matrix[hue_key] = subsetAdata.obs[hue_key]

    # check inputted x values and hue values are in subsetted data
    for i in x_values:
        if i not in list(expression_matrix[x_key].unique()):
            raise ValueError(
                f"{i} is not in adata's x_key {x_key}.\nPossible {x_key}s: {list(adata.obs[x_key].unique())}")
    if hue_key != None:
        for i in hue_values:
            if i not in list(expression_matrix[hue_key].unique()):
                raise ValueError(
                    f"{i} is not in adata's hue_key {hue_key}.\nPossible {hue_key}s: {list(adata.obs[hue_key].unique())}")

    # generate violin plots
    sns.set_theme(context="paper", style="white", )
    for plt_num, gene in enumerate(gene_list):
        plt.subplot(nrows, ncols, plt_num + 1)
        ax = sns.violinplot(
            expression_matrix,
            x=x_key,
            y=gene,
            hue=hue_key,
            density_norm='width',
            width=0.9,
            order=x_values,
            hue_order=hue_values,
            *args, **kwargs
        )
        if hue_key is not None:
            if (plt_num > 0):
                plt.legend([], [], frameon=False)
            else:
                plt.legend(loc='upper right')
        plt.xticks(rotation=90)
        ax.grid(False)
        ax.set(xlabel=None)
        ax.set_title(f"{gene}")
        ax.tick_params(bottom=True, left=True)

    fig.tight_layout()
    if save_path is not None:
        if "/" not in save_path:
            save_path = f"./{save_path}"
        check_path(save_path.rsplit("/", 1)[0])
        logger.info(f"Saving figure to {save_path}")
        plt.savefig(save_path, dpi=dpi, facecolor="white",)
    plt.show()



"""
Loads the lookup tables for an isoform AnnData object.
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""

# external imports
import json
from typing import Tuple, Dict

# scrnatools package imports
from .._configs import configs

logger = configs.create_logger(__name__.split('_', 1)[1])


# -------------------------------------------------------function----------------------------------------------------- #


def load_isoform_lookup_tables(path: str) -> Tuple[Dict[str, str], Dict[str, str], Dict[str, str]]:
    """Loads the lookup tables for an isoform AnnData object.

    Args:
        path (str): The path and file name of the lookup dicts to load. (_ec2tx.json, _ec2g.json, and _inv_map.json all appended to this file name).

    Returns:
        Tuple[Dict[str, str], Dict[str, str], Dict[str, str]]: The equivalence class to transcript, equivalence class to genes, and inverse map dicts.
    """

    with open(f"{path}_ec2tx.json") as d:
        ec2tx = json.load(d)
    with open(f"{path}_ec2g.json") as d:
        ec2g = json.load(d)
    with open(f"{path}_inv_map.json") as d:
        inv_map = json.load(d)
    return ec2tx, ec2g, inv_map



"""
Extracts the cell x gene expression matrix from an AnnData object.
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""
# external package imports
from scipy.sparse import issparse
from pandas import DataFrame
from anndata import AnnData

# scrnatools package imports
from .._configs import configs

logger = configs.create_logger(__name__.split('_', 1)[1])

# -------------------------------------------------------function----------------------------------------------------- #


def get_expression_matrix(
        anndata: AnnData,
        gene_data: str,
) -> DataFrame:
    """Extracts the cell x gene expression matrix from an AnnData object.

    Args:
        anndata (AnnData): The AnnData to pull the expression matrix from.
        gene_data (str): Where to get the expression data from the AnnData object. If a layer from the AnnData['layers'] is passed, that is used, otherwise 'X' or 'raw' can be used.

    Raises:
        ValueError: When 'gene_data' is not one of the layers in 'anndata.layers' or 'X' or 'raw'

    Returns:
        DataFrame: A Pandas DataFrame containing the expression matrix (cells x genes).
    """

    if gene_data == "X":
        # Convert to a dense matrix if needed
        if issparse(anndata.X):
            matrix = DataFrame(
                anndata.X.todense(),
                index=anndata.obs.index,
                columns=anndata.var_names
            )
        else:
            matrix = DataFrame(
                anndata.X,
                index=anndata.obs.index,
                columns=anndata.var_names
            )
    # get a layer from the AnnData if specified
    elif gene_data == "raw":
        raw_data = anndata.raw.to_adata()
        # Convert to a dense matrix if needed
        if issparse(raw_data.X):
            matrix = DataFrame(
                raw_data.X.todense(),
                index=anndata.obs.index,
                columns=anndata.raw.var_names
            )
        else:
            matrix = DataFrame(
                raw_data.X,
                index=anndata.obs.index,
                columns=anndata.raw.var_names
            )
    elif gene_data in anndata.layers:
        if issparse(anndata.layers[gene_data]):
            matrix = DataFrame(
                anndata.layers[gene_data].todense(),
                index=anndata.obs.index,
                columns=anndata.var_names
            )
        else:
            matrix = DataFrame(
                anndata.layers[gene_data],
                index=anndata.obs.index,
                columns=anndata.var_names
            )
    else:
        raise ValueError(
            f"{gene_data} is not 'X', 'raw', or a valid layer name in '{anndata.layers}'"
        )
    return matrix



"""
Creates a UMAP/TSNE plot of a gene's expression from a layer using scanpy.
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""
# external package imports
import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np
import scanpy as sc
import scipy
from typing import List, Tuple
from anndata import AnnData

# scrnatools package imports
from .._configs import configs

logger = configs.create_logger(__name__.split('_', 1)[1])

# -------------------------------------------------------function----------------------------------------------------- #


def gene_embedding(
        adata: AnnData,
        gene_list: List[str],
        layer: str,
        figsize: Tuple[int, int] = (4, 4),
        dpi: int = 80,
        dpi_save: int = 300,
        min_quantile: float = 0.01,
        max_quantile: float = 0.99,
        use_rep="X_umap",
        *args,
        **kwargs,
):
    """Creates a UMAP/TSNE plot of a gene's expression from a layer using scanpy.

    Args:
        adata (AnnData): The dataset containing gene expression data.
        gene_list (List[str]): The list of genes to plot.
        layer (str): The layer containing expression data, or 'X' to use the expression data stored in adata.X.
        figsize (Tuple[int, int], optional): The size of each panel. Defaults to (4, 4).
        dpi (int, optional): The dpi of plots shown inline. Defaults to 80.
        dpi_save (int, optional): The dpi of saved plots. Defaults to 300.
        min_quantile (float, optional):  The quantile of expression for a gene to set the minimum of the colorbar to. Defaults to 0.01.
        max_quantile (float, optional): The quantile of expression for a gene to set the maximum of the colorbar to. Defaults to 0.99.
        use_rep (str, optional): The embedding coordinates in 'adata.obsm' to use. Defaults to "X_umap".

    Raises:
        ValueError: If 'layer' is not a valid key in 'adata.layers'
    """

    # Setup figure
    sc.set_figure_params(figsize=figsize, dpi=dpi,
                         dpi_save=dpi_save, facecolor="white", frameon=False)
    sns.set_context("paper")
    plt.rcParams["axes.grid"] = False
    vmin = []
    vmax = []
    if layer != "X" and layer not in adata.layers.keys():
        raise ValueError(
            f"{layer} is not 'X' or a valid key in 'adata.layers'")
    # Get min/max expression values for the colorbars for each gene
    for gene in gene_list:
        if layer != "X":
            if scipy.sparse.issparse(adata.layers[layer]):
                min = np.quantile(
                    adata[:, gene].layers[layer].todense().tolist(), min_quantile
                )
                max = np.quantile(
                    adata[:, gene].layers[layer].todense().tolist(), max_quantile
                )
                vmin.append(min)
                if max < 1:
                    vmax.append(1)
                else:
                    vmax.append(max)
            else:
                min = np.quantile(adata[:, gene].layers[layer], min_quantile)
                max = np.quantile(adata[:, gene].layers[layer], max_quantile)
                vmin.append(min)
                vmax.append(max)
        else:
            min = np.quantile(adata[:, gene].X, min_quantile)
            max = np.quantile(adata[:, gene].X, max_quantile)
            vmin.append(min)
            vmax.append(max)
    # Plot the data
    adata_copy = adata.copy()
    adata_copy.obsm["X_umap"] = adata.obsm[use_rep]
    if layer == "X":
        sc.pl.umap(
            adata_copy,
            color=gene_list,
            vmin=vmin,
            vmax=vmax,
            *args,
            **kwargs,
        )
    else:
        sc.pl.umap(
            adata_copy,
            color=gene_list,
            layer=layer,
            vmin=vmin,
            vmax=vmax,
            *args,
            **kwargs,
        )



"""
Calculates DE marker genes for data clusters.
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""

# external imports
from anndata import AnnData
from pandas import DataFrame
from typing import Dict
from scvi.model import SCVI

# scrnatools package imports
from .._configs import configs
from .._utils import check_path

logger = configs.create_logger(__name__.split('_', 1)[1])


# -------------------------------------------------------function----------------------------------------------------- #

def cluster_de(
        adata: AnnData,
        model: SCVI,
        cluster_key: str,
        lfc_mean_threshold: int = 0,
        bayes_factor_threshold: int = 3,
        non_zeroes_proportion_threshold: int = 0.1,
        save_path: str = None,
        subset: bool = True,
) -> Dict[str, DataFrame]:
    """Calculates DE marker genes for data clusters.

    Args:
        adata (AnnData): The data to analyze.
        model (SCVI): The scVI model for 'adata'.
        cluster_key (str): The column name of the cluster data in 'adata.obs'.
        lfc_mean_threshold (int, optional): The minimum lfc_mean to filter DE genes on (exclusive). Defaults to 0.
        bayes_factor_threshold (int, optional): The minimum bayes factor to filter de genes on (exclusive). Defaults to 3.
        non_zeroes_proportion_threshold (int, optional): The minimum proportion of cells with non-zero expression filter de genes on (exclusive). Defaults to 0.1.
        save_path (str, optional): The path to save the marker gene lists to. Defaults to None.
        subset (bool, optional): Whether to subset the DE gene list based on thresholds or return all genes in data. Defaults to True.

    Raises:
        ValueError: If 'cluster_key' is not a valid column in 'adata.obs.columns'

    Returns:
        Dict[str, DataFrame]: A dictionary with keys equal to the categories of the cluster column in 'adata.obs' (i.e. cell types) linked to DataFrames of the filtered (or unfiltered but sorted) DE marker genes for that cluster
    """

    if cluster_key not in adata.obs.columns:
        raise ValueError(f"{cluster_key} is not a valis column in 'adata.obs'")
    de_df = model.differential_expression(adata, groupby=cluster_key, )
    cats = adata.obs[cluster_key].cat.categories
    markers = {}
    for i, c in enumerate(cats):
        cid = f"{c} vs Rest"
        cell_type_df = de_df.loc[de_df.comparison == cid]
        if subset:
            cell_type_df = cell_type_df[
                cell_type_df.lfc_mean > lfc_mean_threshold
            ]
            cell_type_df = cell_type_df[
                cell_type_df["bayes_factor"] > bayes_factor_threshold
            ]
            cell_type_df = cell_type_df[
                cell_type_df["non_zeros_proportion1"] > non_zeroes_proportion_threshold
            ]
        markers[c] = cell_type_df
    if save_path is not None:
        check_path(save_path)
        for cluster, marker_genes in markers.items():
            if save_path[-1] == "/":
                save_path = save_path[:-1]
            check_path(save_path)
            marker_genes.to_csv(f"{save_path}/{cluster}.csv")
    return markers



"""
Checks the type of a variable.
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""
from typing import Any

# -------------------------------------------------------function----------------------------------------------------- #


def type_check(var: Any, varname: str, types: Any,):
    """Checks the type of a variable.

    Args:
        var (Any): the variable to check the type of.
        varname (str): the name of the variable.
        types (Any): the type the variable should be.

    Raises:
        TypeError: when 'var' is not one of 'types'.
    """

    if not isinstance(var, types):
        raise TypeError(f"{varname} must be of type {types}")



"""
Saves the lookup tables for an isoform AnnData object.
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""

# external imports
import json
from typing import Dict

# scrnatools package imports
from .._configs import configs

logger = configs.create_logger(__name__.split('_', 1)[1])


# -------------------------------------------------------function----------------------------------------------------- #


def save_isoform_lookup_tables(
        path: str,
        ec2tx: Dict[str, str],
        ec2g: Dict[str, str],
        inv_map: Dict[str, str],
):
    """Saves the lookup tables for an isoform AnnData object.

    Args:
        path (str): The path and file name to save the lookup dicts to. (_ec2tx.json, _ec2g.json, and _inv_map.json all appended to this file name).
        ec2tx (Dict[str, str]): The equivalence class to transcript dict to save.
        ec2g (Dict[str, str]): The equivalence class to gene dict to save.
        inv_map (Dict[str, str]): The gene to equivalence class dict to save.
    """

    with open(f"{path}_ec2tx.json", "w") as f:
        json_file = json.dumps(ec2tx)
        f.write(json_file)
    with open(f"{path}_ec2g.json", "w") as f:
        json_file = json.dumps(ec2g)
        f.write(json_file)
    with open(f"{path}_inv_map.json", "w") as f:
        json_file = json.dumps(inv_map)
        f.write(json_file)



"""
Reads in a kallisto bustools isoform matrix.
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""

# external imports
import scanpy as sc
import pandas as pd
from pandas import DataFrame
from typing import Tuple
from anndata import AnnData
import numpy as np

# scrnatools package imports
from .._configs import configs

logger = configs.create_logger(__name__.split('_', 1)[1])


# -------------------------------------------------------function----------------------------------------------------- #


def read_kallisto_tcc_matrix(
        sample_path: str,
) -> Tuple[AnnData, DataFrame]:
    """Reads in a kallisto bustools isoform matrix.

    Args:
        sample_path (str): The path to and sample name for the mtx, barcodes, and ec files from kallisto bustools output.

    Returns:
        Tuple[AnnData, DataFrame]: An AnnData with the tcc data and the ec matrix.
    """

    logger.info(f"Reading sample: {sample_path}")
    data = sc.read_mtx(f"{sample_path}.mtx")
    data.obs_names = pd.read_csv(
        f"{sample_path}.barcodes.txt",
        header=None
    ).values.T[0]
    ec = pd.read_csv(f"{sample_path}.ec.txt", sep='\t', header=None)
    data.var_names = np.array(np.arange(data.shape[1]), dtype=str)
    return data, ec



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="scrnatools",
    version="1.12",
    author="Joe Germino",
    author_email="joe.germino@ucsf.edu",
    description="Tools for single cell RNA sequencing pipelines",
    long_description="scrnatools provides helper plotting and analysis functions built on top of the scanpy and scVI packages to perform common tasks in scRNA-seq analysis pipelines. Extensive description of general best practices and detailed default pipeline examples can be found in the scanpy and scvi-tools documentation. For an example walkthrough of a typical scRNA-seq pipeline and the added functionality provided by scrnatools see the jupyter notebooks provided at https://github.com/j-germino/scrnatools-git/tree/main/examples. These can be uploaded to Google drive and run within a Google colab environemnt, using a GPU session. API documentation is available at: https://scrnatools-git.readthedocs.io/en/latest/.",
    url="https://github.com/j-germino/sc-rna-tools",
    packages=setuptools.find_packages(),
    install_requires=[
        "scanpy",
        "scrublet",
        "scvi-tools",
        "matplotlib",
        "pandas",
        "numpy",
        "seaborn",
        "scikit-misc",
        "leidenalg",
        "anndata",
        "scipy",
        "starlette"
    ],
    python_requires='>=3,',
)



"""
Checks if a path exists and creates it if it doesn't.
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""
# external package imports
import os

# ----------------------function---------------------- #


def check_path(path: str):
    """Checks if a path exists and creates it if it doesn't.

    Args:
        path (str): The string of the path to check/create.
    """

    if not os.path.exists(path):
        os.makedirs(path)



"""
Filters cells based on gene number, total counts, and % mitochondrial reads.
From scrnatools package

Created on Mon Jan 10 15:57:46 2022

@author: joe germino (joe.germino@ucsf.edu)
"""

# external imports
from anndata import AnnData
from typing import Tuple

# scrnatools package imports
from .._configs import configs

logger = configs.create_logger(__name__.split('_', 1)[1])

# -------------------------------------------------------function----------------------------------------------------- #


def filter_cells(
        adata: AnnData,
        genes_thresholds: Tuple[int, int],
        counts_thresholds: Tuple[int, int],
        mt_threshold: int = 10,
) -> AnnData:
    """Filters cells based on gene number, total counts, and % mitochondrial reads.

    Args:
        adata (AnnData): The AnnData with the data to filter.
        genes_thresholds (Tuple[int, int]): A Tuple of thresholds for the number of genes per cell with 'gene_thresholds[0]' being the lower bound and 'gene_thresholds[1]' being the upper bound (both exclusive).
        counts_thresholds (Tuple[int, int]): A Tuple of thresholds for the number of total counts per cell with 'count_thresholds[0]' being the lower bound and 'count_thresholds[1]' being the upper bound (both exclusive).
        mt_threshold (int, optional): The maximum percent mitochondrial reads per cell. Defaults to 10.

    Returns:
        AnnData:  The dataset filtered on cells that don't pass the thresholds.
    """

    logger.info(f"Number of cells before QC filtering: {len(adata.obs)}")
    filtered_adata = adata[adata.obs.pct_counts_mt < mt_threshold].copy()
    filtered_adata = filtered_adata[
        filtered_adata.obs.total_counts < counts_thresholds[1]
    ]
    filtered_adata = filtered_adata[
        filtered_adata.obs.total_counts > counts_thresholds[0]
    ]
    filtered_adata = filtered_adata[
        filtered_adata.obs.n_genes_by_counts < genes_thresholds[1]
    ]
    filtered_adata = filtered_adata[
        filtered_adata.obs.n_genes_by_counts > genes_thresholds[0]
    ].copy()
    logger.info(
        f"Number of cells after QC filtering: {len(filtered_adata.obs)}")
    return filtered_adata
